from setuptools import setup

setup(
    name='msglink',
    version='1.5.6',
    packages=['msglink'],
    url='',
    license='LGPL v3',
    author='Robert Brünig',
    author_email='',
    description='msglink client library',
    package_data={"msglink": ["py.typed"]},
)
